package ru.dbotthepony.mc.otm.capability.matter;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.nbt.CompoundTag;
import net.minecraftforge.common.util.INBTSerializable;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class MatterHandlerCapability implements IMatterHandler, INBTSerializable<CompoundTag> {
	protected final Runnable listener;
	protected Fraction stored = Fraction.ZERO;
	protected MatterDirection direction;
	protected Fraction max_storage;

	protected Fraction max_receive;
	protected Fraction max_extract;

	public MatterHandlerCapability(
		@Nullable Runnable listener,
		MatterDirection direction,
		Fraction max_storage
	) {
		this.listener = listener;
		this.direction = direction;
		this.max_storage = max_storage;
	}

	public MatterHandlerCapability(
		@Nullable Runnable listener,
		MatterDirection direction,
		Fraction max_storage,
		Fraction max_receive,
		Fraction max_extract
	) {
		this(listener, direction, max_storage);

		this.max_receive = max_receive;
		this.max_extract = max_extract;
	}

	private LazyOptional<IMatterHandler> handler = LazyOptional.of(() -> this);

	public void invalidate() {
		handler.invalidate();
	}

	public void revive() {
		handler = LazyOptional.of(() -> this);
	}

	public LazyOptional<IMatterHandler> get() {
		return handler;
	}

	@Nonnull
	@Override
	public Fraction getStoredMatter() {
		return stored;
	}

	@Nonnull
	@Override
	public Fraction getMaxStoredMatter() {
		return max_storage;
	}

	public boolean canReceiveAll(Fraction value) {
		return max_storage.compareTo(value) >= 0 && stored.plus(value).compareTo(max_storage) <= 0;
	}

	@Nonnull
	@Override
	public Fraction receiveMatterOuter(Fraction howMuch, boolean simulate) {
		if (getDirection() == MatterDirection.EXTRACT)
			return Fraction.ZERO;

		return receiveMatterInner(howMuch, simulate);
	}

	@Nonnull
	@Override
	public Fraction receiveMatterInner(Fraction howMuch, boolean simulate) {
		Fraction new_matter;

		if (max_receive == null) {
			new_matter = stored.plus(howMuch).min(max_storage);
		} else {
			new_matter = stored.plus(howMuch.min(max_receive)).min(max_storage);
		}

		Fraction diff = new_matter.minus(stored);

		if (!simulate && new_matter.compareTo(stored) != 0) {
			stored = new_matter;

			if (listener != null) {
				listener.run();
			}
		}

		return diff;
	}

	@Nonnull
	@Override
	public Fraction extractMatterOuter(Fraction howMuch, boolean simulate) {
		if (getDirection() == MatterDirection.RECEIVE)
			return Fraction.ZERO;

		return extractMatterInner(howMuch, simulate);
	}

	@Nonnull
	@Override
	public Fraction extractMatterInner(Fraction howMuch, boolean simulate) {
		Fraction new_matter;

		if (max_receive == null) {
			new_matter = stored.minus(howMuch).moreThanZero();
		} else {
			new_matter = stored.minus(howMuch.min(max_receive)).moreThanZero();
		}

		Fraction diff = stored.minus(new_matter);

		if (!simulate && new_matter.compareTo(stored) != 0) {
			stored = new_matter;

			if (listener != null) {
				listener.run();
			}
		}

		return diff;
	}

	@Nonnull
	@Override
	public MatterDirection getDirection() {
		return direction;
	}

	@Override
	public CompoundTag serializeNBT() {
		CompoundTag tag = new CompoundTag();
		tag.put("stored", stored.serializeNBT());
		tag.put("max_storage", max_storage.serializeNBT());

		if (max_receive != null)
			tag.put("max_receive", max_receive.serializeNBT());

		if (max_extract != null)
			tag.put("max_extract", max_extract.serializeNBT());

		return tag;
	}

	@Override
	public void deserializeNBT(CompoundTag tag) {
		if (tag.contains("stored"))
			stored = Fraction.deserializeNBT(tag.get("stored"));

		if (tag.contains("max_storage"))
			max_storage = Fraction.deserializeNBT(tag.get("max_storage"));

		if (tag.contains("max_receive"))
			max_receive = Fraction.deserializeNBT(tag.get("max_receive"));
		else
			max_receive = null;

		if (tag.contains("max_extract"))
			max_extract = Fraction.deserializeNBT(tag.get("max_extract"));
		else
			max_extract = null;
	}
}
