package ru.dbotthepony.mc.otm.container;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.world.Container;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.items.IItemHandler;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.util.Arrays;
import java.util.Iterator;
import java.util.function.Consumer;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class MatteryContainer implements Container, Iterable<ItemStack> {
	protected final Runnable watcher;
	protected int ignore_change_notifications = 0;
	protected final int size;
	protected final ItemStack[] slots;
	protected final ItemStack[] tracked_slots;

	public MatteryContainer(Runnable watcher, int size) {
		this.size = size;

		slots = new ItemStack[this.size];
		tracked_slots = new ItemStack[this.size];
		Arrays.fill(slots, ItemStack.EMPTY);
		Arrays.fill(tracked_slots, ItemStack.EMPTY);

		this.watcher = watcher;
	}

	@SuppressWarnings("unused")
	public void startIgnore() {
		ignore_change_notifications++;
	}

	@SuppressWarnings("unused")
	public void stopIgnore() {
		ignore_change_notifications--;
	}

	public MatteryContainer of(@Nullable CompoundTag tag) {
		if (tag == null)
			return this;

		MatteryContainer container = MatteryContainer.of(watcher, tag, getContainerSize());

		if (handler != null)
			container.handler = handler.cloneOf(container);

		return container;
	}

	public void deserializeNBT(@Nullable CompoundTag tag) {
		ignore_change_notifications++;

		for (int i = 0; i < getContainerSize(); i++) {
			setItem(i, ItemStack.EMPTY);
		}

		if (tag == null) {
			ignore_change_notifications--;
			return;
		}

		// нам не интересен размер

		if (tag.get("items") instanceof ListTag list) {
			for (int i = 0; i < Math.min(list.size(), getContainerSize()); i++) {
				if (list.get(i) instanceof CompoundTag get_tag) {
					setItem(i, ItemStack.of(get_tag));
				}
			}
		}

		ignore_change_notifications--;
		setChanged();
	}

	public MatteryContainer of(@Nullable Tag tag) {
		if (tag == null)
			return this;

		if (tag instanceof CompoundTag compound)
			return of(watcher, compound, getContainerSize());

		return this;
	}

	public void deserializeNBT(@Nullable Tag tag) {
		if (tag == null)
			return;

		if (tag instanceof CompoundTag compound) {
			deserializeNBT(compound);
		}
	}

	public ContainerIteratorItemStack stackIterator() {
		return new ContainerIteratorItemStack(this);
	}

	@SuppressWarnings("unused")
	public void consumeNonEmpty(Consumer<ItemStack> consumer) {
		stackIterator().consume(consumer);
	}

	public <T> ContainerIteratorCapability<T> capabilityIterator(Capability<T> cap) {
		return new ContainerIteratorCapability<>(this, cap);
	}

	public <T> void consumeCapability(Capability<T> cap, Consumer<T> consumer) {
		capabilityIterator(cap).consume(consumer);
	}

	public void setChanged(int slot, ItemStack new_state, ItemStack old_state) {
		if (ignore_change_notifications == 0)
			watcher.run();
	}

	public CompoundTag serializeNBT() {
		CompoundTag tag = new CompoundTag();
		tag.putInt("size", getContainerSize());
		ListTag list = new ListTag();
		tag.put("items", list);

		for (int i = 0; i < getContainerSize(); i++)
			list.add(getItem(i).serializeNBT());

		return tag;
	}

	public static MatteryContainer of(Runnable watcher, @Nullable CompoundTag tag, int fallback_size) {
		if (tag == null)
			return new MatteryContainer(watcher, fallback_size);

		MatteryContainer container;

		if (tag.contains("size"))
			container = new MatteryContainer(watcher, tag.getInt("size"));
		else
			container = new MatteryContainer(watcher, fallback_size);

		if (tag.get("items") instanceof ListTag list)
			for (int i = 0; i < Math.min(list.size(), container.getContainerSize()); i++)
				if (list.get(i) instanceof CompoundTag get_tag)
					container.setItem(i, ItemStack.of(get_tag));

		return container;
	}

	@SuppressWarnings("unused")
	public boolean hasEmptySlot() {
		for (int i = 0; i < getContainerSize(); i++) {
			if (getItem(i).isEmpty())
				return true;
		}

		return false;
	}

	protected MatteryContainerHandler handler;

	public MatteryContainerHandler handler(MatteryContainerInsertValidator insert_validator, MatteryContainerExtractValidator extract_validator) {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this, insert_validator, extract_validator);
	}

	public MatteryContainerHandler handler(MatteryContainerInsertValidator insert_validator) {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this, insert_validator);
	}

	public MatteryContainerHandler handler() {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this);
	}

	public int getMaxStackSize(int slot) {
		return getMaxStackSize();
	}

	public ItemStack addItem(ItemStack stack, int start, int end, boolean simulate) {
		if (stack.isEmpty() || start < 0 || end > size || start >= end)
			return ItemStack.EMPTY;

		final var copy = stack.copy();

		// двигаем в одинаковые слоты
		for (int slot = start; slot < end; slot++) {
			if (ItemStack.isSameItemSameTags(slots[slot], copy)) {
				var this_stack = slots[slot];
				int slot_limit = Math.min(getMaxStackSize(slot), this_stack.getMaxStackSize());

				if (this_stack.getCount() < slot_limit) {
					int new_count = Math.min(this_stack.getCount() + copy.getCount(), slot_limit);
					int diff = new_count - this_stack.getCount();

					if (!simulate) {
						var old = this_stack.copy();
						this_stack.setCount(new_count);
						tracked_slots[slot] = this_stack.copy();
						setChanged(slot, this_stack, old);
					}

					copy.shrink(diff);

					if (copy.isEmpty()) {
						return copy;
					}
				}
			}
		}

		// двигаем в пустые слоты
		for (int slot = start; slot < end; slot++) {
			if (slots[slot].isEmpty()) {
				int diff = Math.min(copy.getCount(), Math.min(getMaxStackSize(slot), copy.getMaxStackSize()));

				if (!simulate) {
					var put_copy = copy.copy();
					put_copy.setCount(diff);
					slots[slot] = put_copy;
					tracked_slots[slot] = put_copy.copy();
					setChanged(slot, put_copy, ItemStack.EMPTY);
				}

				copy.shrink(diff);

				if (copy.isEmpty()) {
					return copy;
				}
			}
		}

		return copy;
	}

	@SuppressWarnings("unused")
	public ItemStack addItem(ItemStack stack, boolean simulate) {
		return addItem(stack, 0, size, simulate);
	}

	@SuppressWarnings("unused")
	public ItemStack addItem(ItemStack stack) {
		return addItem(stack, 0, size, false);
	}

	public boolean fullyAddItem(ItemStack stack) {
		if (!addItem(stack, 0, size, true).isEmpty()) {
			return false;
		}

		return addItem(stack, 0, size, false).isEmpty();
	}

	@SuppressWarnings("unused")
	public boolean fullyAddItem(ItemStack stack, int start, int end) {
		if (!addItem(stack, start, end, true).isEmpty()) {
			return false;
		}

		return addItem(stack, start, end, false).isEmpty();
	}

	@Override
	public int getContainerSize() {
		return size;
	}

	@Override
	public boolean isEmpty() {
		for (var stack : slots)
			if (!stack.isEmpty())
				return false;

		return true;
	}

	@Override
	public ItemStack getItem(int slot) {
		return slots[slot].isEmpty() ? ItemStack.EMPTY : slots[slot];
	}

	@Override
	public ItemStack removeItem(int slot, int amount) {
		if (amount <= 0 || slot < 0 || slot >= size || slots[slot].isEmpty())
			return ItemStack.EMPTY;

		var old = slots[slot].copy();
		var split = slots[slot].split(amount);
		tracked_slots[slot] = slots[slot].copy();
		setChanged(slot, slots[slot].isEmpty() ? ItemStack.EMPTY : slots[slot], old);

		return split;
	}

	@Override
	public ItemStack removeItemNoUpdate(int slot) {
		var old = slots[slot];
		slots[slot] = ItemStack.EMPTY;
		tracked_slots[slot] = ItemStack.EMPTY;
		return old;
	}

	@Override
	public void setItem(int slot, ItemStack stack) {
		if (slots[slot].isEmpty() && stack.isEmpty() || stack == slots[slot])
			return;

		var old = slots[slot];
		slots[slot] = stack;
		tracked_slots[slot] = stack.copy();
		setChanged(slot, stack, old);
	}

	@Override
	public void setChanged() {
		for (int slot = 0; slot < size; slot++) {
			if (!ItemStack.isSameItemSameTags(slots[slot], tracked_slots[slot])) {
				setChanged(slot, slots[slot], tracked_slots[slot]);
				tracked_slots[slot] = slots[slot].copy();
				// mojang соси))0)0))0)))))0)
			}
		}
	}

	@Override
	public boolean stillValid(Player player) {
		return true;
	}

	@Override
	public void clearContent() {
		for (int slot = 0; slot < size; slot++) {
			if (!slots[slot].isEmpty()) {
				setChanged(slot, ItemStack.EMPTY, slots[slot]);
			}
		}

		Arrays.fill(tracked_slots, ItemStack.EMPTY);
	}

	@Override
	public Iterator<ItemStack> iterator() {
		return Arrays.stream(slots).iterator();
	}
}
