package ru.dbotthepony.mc.otm.block.entity

import javax.annotation.ParametersAreNonnullByDefault
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.capability.matter.IPatternStorage
import ru.dbotthepony.mc.otm.container.MatteryContainer
import java.lang.Runnable
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import ru.dbotthepony.mc.otm.block.BlockPatternStorage
import net.minecraft.nbt.CompoundTag
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import ru.dbotthepony.mc.otm.menu.PatternStorageMenu
import net.minecraft.MethodsReturnNonnullByDefault
import net.minecraft.core.Direction
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.capability.matter.PatternInsertStatus
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set
import java.util.ArrayList

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
class BlockEntityPatternStorage(p_155229_: BlockPos, p_155230_: BlockState) :
	BlockEntityMattery(Registry.BlockEntities.PATTERN_STORAGE, p_155229_, p_155230_), IMatterGraphNode, IPatternStorage {

	private val node = Graph6Node<IMatterGraphNode>(this)
	private val resolverPatterns = LazyOptional.of { this }
	private val resolverNode = LazyOptional.of { this }

	@JvmField
	val patterns: MatteryContainer = object : MatteryContainer(Runnable { this.setChanged() }, 8) {
		override fun setChanged(slot: Int, new_state: ItemStack, old_state: ItemStack) {
			val grid = node.graph as MatterNetworkGraph?

			if (grid != null && !ItemStack.isSameItemSameTags(new_state, old_state)) {
				if (!old_state.isEmpty) {
					old_state.getCapability(MatteryCapability.PATTERN).ifPresent { cap: IPatternStorage ->
						cap.storedPatterns.forEach { state: PatternState? -> grid.onPatternRemoved(state!!) }
					}
				}

				if (!new_state.isEmpty) {
					new_state.getCapability(MatteryCapability.PATTERN).ifPresent { cap: IPatternStorage ->
						cap.storedPatterns.forEach { state: PatternState? -> grid.onPatternAdded(state!!) }
					}
				}

				updateBlockstate()
			}

			super.setChanged(slot, new_state, old_state)
		}

		override fun getMaxStackSize(): Int {
			return 1
		}
	}

	private fun updateBlockstate() {
		val level = level ?: return

		var state = blockState

		for (i in 0..7) {
			state = state.setValue(
				BlockPatternStorage.PATTERN_STORAGE_DISKS_PROPS[i],
				patterns.getItem(i).getCapability(MatteryCapability.PATTERN).isPresent
			)
		}

		if (state !== blockState) {
			level.setBlock(blockPos, state, Block.UPDATE_CLIENTS)
		}
	}

	private val resolverItem =
		patterns.handler { slot: Int, stack: ItemStack -> stack.getCapability(MatteryCapability.PATTERN).isPresent }

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["patterns"] = patterns.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("patterns") {
			patterns.deserializeNBT(it)
		}
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	override fun getPatternHandler(): IPatternStorage {
		return this
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverItem.invalidate()
		node.destroy(::MatterNetworkGraph)
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverItem.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap === MatteryCapability.PATTERN) return resolverPatterns.cast()
			if (cap === MatteryCapability.MATTER_CELL) return resolverNode.cast()
			if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) return resolverItem.get().cast()
		}

		return super.getCapability(cap, side)
	}

	override fun getDefaultDisplayName(): Component {
		return NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return PatternStorageMenu(containerID, inventory, this)
	}

	override fun getStoredPatterns(): Collection<PatternState> {
		val list = ArrayList<PatternState>()
		patterns.consumeCapability(MatteryCapability.PATTERN) { capability: IPatternStorage -> list.addAll(capability.storedPatterns) }
		return list
	}

	override fun getCapacity(): Int {
		var stored = 0L

		for (pattern in patterns.capabilityIterator(MatteryCapability.PATTERN))
			stored += pattern.capacity.toLong()

		return if (stored > Int.MAX_VALUE) Int.MAX_VALUE else stored.toInt()
	}

	override fun getStored(): Int {
		var stored = 0L

		for (pattern in patterns.capabilityIterator(MatteryCapability.PATTERN))
			stored += pattern.stored.toLong()

		return if (stored > Int.MAX_VALUE) Int.MAX_VALUE else stored.toInt()
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	override fun insertPattern(pattern: PatternState, only_update: Boolean, simulate: Boolean): PatternInsertStatus {
		for (storage in patterns.capabilityIterator(MatteryCapability.PATTERN)) {
			val status = storage.insertPattern(pattern, only_update, simulate)

			if (status.status() != PatternInsertStatus.Status.FAIL) {
				if (!simulate) {
					setChanged()

					val graph = node.graph as MatterNetworkGraph?

					if (graph != null) {
						if (status.status() == PatternInsertStatus.Status.INSERTED) {
							graph.onPatternAdded(status.new_state()!!)
						} else {
							graph.onPatternUpdated(status.new_state()!!, status.old_state()!!)
						}
					}
				}

				return status
			}
		}

		return PatternInsertStatus()
	}

	companion object {
		private val NAME = TranslatableComponent("block.overdrive_that_matters.pattern_storage")
	}
}